@@ -1,22 +1,22 @@
 /* Copyright (c) 2006 Trusted Computer Solutions, Inc. */
-
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/poll.h>
-#include <sys/stat.h>
-#include <sys/un.h>
 #include <errno.h>
+#include <poll.h>
+#include <signal.h>
 #include <stdint.h>
 #include <stdio.h>
-#include <unistd.h>
 #include <stdlib.h>
-#include <signal.h>
 #include <string.h>
 #include <syslog.h>
+#include <unistd.h>
 #include <selinux/selinux.h>
-#include <sys/types.h>
 #include <sys/capability.h>
 #include <sys/resource.h>
+#include <sys/socket.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/uio.h>
+#include <sys/un.h>
+#include "mcstrans.h"
 
 #ifdef UNUSED
 #elif defined(__GNUC__)
@@ -32,6 +32,7 @@
 #define SETRANS_INIT			1
 #define RAW_TO_TRANS_CONTEXT		2
 #define TRANS_TO_RAW_CONTEXT		3
+#define RAW_CONTEXT_TO_COLOR		4
 #define MAX_DATA_BUF			4096
 #define MAX_DESCRIPTORS			8192
 
@@ -47,6 +48,10 @@ extern void finish_context_translations(void);
 extern int trans_context(const security_context_t, security_context_t *);
 extern int untrans_context(const security_context_t, security_context_t *);
 
+extern int init_colors(void);
+extern void finish_context_colors(void);
+extern int raw_color(const security_context_t, char **);
+
 #define SETRANSD_PATHNAME "/sbin/mcstransd"
 
 /* name of program (for error messages) */
@@ -59,53 +64,23 @@ static void cleanup_exit(int ret) __attribute__ ((noreturn));
 static void
 cleanup_exit(int ret) 
 {
+	finish_context_colors();
 	finish_context_translations();
 	if (sockfd >=0)
 		(void)unlink(SETRANS_UNIX_SOCKET);
+
+	log_debug("%s\n", "cleanup_exit");
+
 	exit(ret);
 }
 
 static void clean_exit(void);
 static  __attribute__((noreturn)) void clean_exit(void)
 {
+	log_debug("%s\n", "clean_exit");
 	cleanup_exit(0);
 }
 
-/*
- * Convert raw label portion of a security context to translated label
- * Returns:  0 on success, 1 on failure
- */
-static int
-raw_to_trans_context(char *in, char **out, char *UNUSED(pcon))
-{
-
-	*out = NULL; 
-
-	/* TODO: Check if MLS clearance (in "pcon") dominates the MLS label
-	 * (in "in").
-	 */
-
-	
-	return 	trans_context(in, out);
-}
-
-
-/*
- * Convert translated label of a security context to raw label
- * Returns:  0 on success, 1 on failure
- */
-static int
-trans_to_raw_context(char *in, char **out, char *UNUSED(pcon))
-{
-	*out = NULL;
-	
-	/* TODO: Check if MLS clearance (in "pcon") dominates the MLS label
-	 * (in "in").
-	 */
-
-	return untrans_context(in, out);
-}
-
 static int
 send_response(int fd, uint32_t function, char *data, int32_t ret_val)
 {
@@ -115,7 +90,7 @@ send_response(int fd, uint32_t function, char *data, int32_t ret_val)
 	ssize_t count;
 
 	if (!data)
-		data = "";
+		data = (char *)"";
 
 	data_size = strlen(data) + 1;
 
@@ -126,9 +101,6 @@ send_response(int fd, uint32_t function, char *data, int32_t ret_val)
 	resp_hdr[2].iov_base = &ret_val;
 	resp_hdr[2].iov_len = sizeof(ret_val);
 
-	log_debug("Response header func=%d size=%u ret_val=%d",
-	          function, data_size, ret_val);
-
 	while (((count = writev(fd, resp_hdr, 3)) < 0) && (errno == EINTR));
 	if (count != (sizeof(function) + sizeof(data_size) + sizeof(ret_val))) {
 		syslog(LOG_ERR, "Failed to write response header");
@@ -138,19 +110,17 @@ send_response(int fd, uint32_t function, char *data, int32_t ret_val)
 	resp_data.iov_base = data;
 	resp_data.iov_len = data_size;
 
-	log_debug("Response data=%s", data);
-
 	while (((count = writev(fd, &resp_data, 1)) < 0) && (errno == EINTR));
 	if (count < 0 || (size_t)count != data_size) {
 		syslog(LOG_ERR, "Failed to write response data");
 		return -1;
 	}
 
-	return 0;
+	return ret_val;
 }
 
 static int
-get_peer_con(int fd, char **peercon)
+get_peer_pid(int fd, pid_t *pid)
 {
 	int ret;
 	socklen_t size = sizeof(struct ucred);
@@ -162,16 +132,11 @@ get_peer_con(int fd, char **peercon)
 		syslog(LOG_ERR, "Failed to get PID of client process");
 		return -1;
 	}
-	ret = getpidcon_raw(peercred.pid, peercon);
-	if (ret) {
-		syslog(LOG_ERR, 
-			"Failed to get context of client process (pid=%u)",
-			peercred.pid);
-		return -1;
-	}
-	return 0;
+	*pid = peercred.pid;
+	return ret;
 }
 
+
 static int
 process_request(int fd, uint32_t function, char *data1, char *UNUSED(data2))
 {
@@ -180,34 +145,44 @@ process_request(int fd, uint32_t function, char *data1, char *UNUSED(data2))
 	char *peercon = NULL;
 	int ret;
 
-	log_debug("Processing request func=%d data1=%s"
-	          , function, data1);
+	ret = getpeercon_raw(fd, &peercon);
+	if (ret < 0)
+		return ret;
 
+	/* TODO: Check if MLS clearance (in peercon) dominates the MLS label
+	 * (in the request input).
+	 */
+  
 	switch (function) {
 	case SETRANS_INIT:
 		result = 0;
 		ret = send_response(fd, function, NULL, result);
 		break;
 	case RAW_TO_TRANS_CONTEXT:
-		ret = get_peer_con(fd, &peercon);
-		if (ret)
-			return ret;
-		result = raw_to_trans_context(data1, &out, peercon);
+		result = trans_context(data1, &out);
 		ret = send_response(fd, function, out, result);
 		break;
 	case TRANS_TO_RAW_CONTEXT:
-		ret = get_peer_con(fd, &peercon);
-		if (ret)
-			return ret;
-		result = trans_to_raw_context(data1, &out, peercon);
+		result = untrans_context(data1, &out);
+		ret = send_response(fd, function, out, result);
+		break;
+	case RAW_CONTEXT_TO_COLOR:
+		result = raw_color(data1, &out);
 		ret = send_response(fd, function, out, result);
 		break;
 	default:
-		syslog(LOG_ERR, "Invalid request func=%d", function);
+		result = -1;
 		ret = -1;
 		break;
 	}
 
+	if (result) {
+		pid_t pid = 0;
+		get_peer_pid(fd, &pid);
+		syslog(LOG_ERR, "Invalid request func=%d from=%u",
+		       function, pid);
+	}
+
 	free(out);
 	freecon(peercon);
 
@@ -236,36 +211,32 @@ service_request(int fd)
 
 	while (((count = readv(fd, req_hdr, 3)) < 0) && (errno == EINTR));
 	if (count <= 0) {
-		log_debug("Received no data for fd (%d)", fd);
 		return 1;
 	}
 	if (count != (sizeof(function) + sizeof(data1_size) +
 	              sizeof(data2_size) )) {
-		log_debug("Failed to read request header %d != %u",(int)count,
+		log_debug("Failed to read request header %d != %u\n",(int)count,
 			(unsigned)(sizeof(function) + sizeof(data1_size) +
                       sizeof(data2_size) ));
 		return -1;
 	}
 
-	log_debug("Received header func=%d size1=%u size2=%u",
-	          function, data1_size, data2_size);
-
 	if (!data1_size || !data2_size || data1_size > MAX_DATA_BUF ||
 						data2_size > MAX_DATA_BUF ) {
-		log_debug("Header invalid data1_size=%u data2_size=%u",
+		log_debug("Header invalid data1_size=%u data2_size=%u\n",
 		        data1_size, data2_size);
 		return -1;
 	}
 
 	data1 = malloc(data1_size);
 	if (!data1) {
-		log_debug("Could not allocate %d bytes", data1_size);
+		log_debug("Could not allocate %d bytes\n", data1_size);
 		return -1; 
 	}
 	data2 = malloc(data2_size);
 	if (!data2) {
 		free(data1);
-		log_debug("Could not allocate %d bytes", data2_size);
+		log_debug("Could not allocate %d bytes\n", data2_size);
 		return -1;
 	}
 
@@ -279,12 +250,10 @@ service_request(int fd)
 	    data1[data1_size - 1] != '\0' || data2[data2_size - 1] != '\0') {
 		free(data1);
 		free(data2);
-		log_debug("Failed to read request data (%d)", (int)count);
+		log_debug("Failed to read request data (%d)\n", (int)count);
 		return -1;
 	}
 
-	log_debug("Received data1=%s data2=%s", data1, data2);
-
 	ret = process_request(fd, function, data1, data2);
 
 	free(data1);
@@ -349,9 +318,6 @@ process_events(struct pollfd **ufds, int *nfds)
 		short revents = (*ufds)[ii].revents;
 		int connfd = (*ufds)[ii].fd;
 
-		log_debug("Events returned for fd (%d) are (%x)",
-		          connfd, revents);
-
 		if (revents & (POLLIN | POLLPRI)) {
 			if (connfd == sockfd) {
 
@@ -361,12 +327,9 @@ process_events(struct pollfd **ufds, int *nfds)
 					return -1;
 				}
 
-				log_debug("Accepted connection with fd (%d)",
-					 connfd);
-
 				if (add_pollfd(ufds, nfds, connfd)) {
 					syslog(LOG_ERR,
-					  "Failed to add fd (%d) to poll list",
+					  "Failed to add fd (%d) to poll list\n",
 						connfd);
 					return -1;
 				}
@@ -376,11 +339,8 @@ process_events(struct pollfd **ufds, int *nfds)
 					if (ret < 0) {
 						syslog(LOG_ERR,
 							"Servicing of request "
-							"failed for fd (%d)",
+							"failed for fd (%d)\n",
 							connfd);
-					} else {
-						log_debug("Closing fd (%d)",
-							  connfd);
 					}
 					/* Setup pollfd for deletion later. */
 					(*ufds)[ii].fd = -1;
@@ -392,7 +352,7 @@ process_events(struct pollfd **ufds, int *nfds)
 			revents = revents & ~(POLLIN | POLLPRI);
 		}
 		if (revents & POLLHUP) {
-			log_debug("The connection with fd (%d) hung up",
+			log_debug("The connection with fd (%d) hung up\n",
 				connfd);
 
 			/* Set the pollfd up for deletion later. */
@@ -403,7 +363,7 @@ process_events(struct pollfd **ufds, int *nfds)
 		}
 		if (revents) {
 			syslog(LOG_ERR, "Unknown/error events (%x) encountered"
-					" for fd (%d)", revents, connfd);
+					" for fd (%d)\n", revents, connfd);
 
 			/* Set the pollfd up for deletion later. */
 			(*ufds)[ii].fd = -1;
@@ -426,10 +386,9 @@ static void
 process_connections(void)
 {
 	int ret = 0;
-	struct pollfd *ufds;
-	int nfds;
+	int nfds = 1;
 
-	ufds = (struct pollfd *)malloc(sizeof(struct pollfd));
+	struct pollfd *ufds = (struct pollfd *)malloc(sizeof(struct pollfd));
 	if (!ufds) {
 		syslog(LOG_ERR, "Failed to allocate a pollfd");
 		cleanup_exit(1);
@@ -437,17 +396,23 @@ process_connections(void)
 	ufds[0].fd = sockfd;
 	ufds[0].events = POLLIN|POLLPRI;
 	ufds[0].revents = 0;
-	nfds = 1;
 
 	while (1) {
 		if (restart_daemon) {
 			syslog(LOG_NOTICE, "Reload Translations");
+			finish_context_colors();
 			finish_context_translations();
-			init_translations();
+			if (init_translations()) {
+				syslog(LOG_ERR, "Failed to initialize label translations");
+				cleanup_exit(1);
+			}
+			if (init_colors()) {
+				syslog(LOG_ERR, "Failed to initialize color translations");
+				syslog(LOG_ERR, "No color information will be available");
+			}
 			restart_daemon = 0;
 		}
 
-		log_debug("About to poll for %d fds", nfds);
 		ret = poll(ufds, nfds, -1);
 		if (ret < 0) {
 			if (errno == EINTR) {
@@ -491,6 +456,10 @@ initialize(void)
 		syslog(LOG_ERR, "Failed to initialize label translations");
 		cleanup_exit(1);
 	}
+	if (init_colors()) {
+		syslog(LOG_ERR, "Failed to initialize color translations");
+		syslog(LOG_ERR, "No color information will be available");
+	}
 
 	/* the socket will be unlinked when the daemon terminates */
 	act.sa_handler = sigterm_handler;
@@ -563,14 +532,36 @@ void dropprivs(void)
 	if (cap_set_proc(new_caps)) {
 		syslog(LOG_ERR, "Error dropping capabilities, aborting: %s\n",
 			 strerror(errno));
-		exit(-1);
+		cleanup_exit(-1);
 	}
 	cap_free(new_caps);
 }
 
-int
-main(int UNUSED(argc), char *argv[])
+static void usage(char *program)
 {
+	printf("%s [-f] [-h] \n", program);
+}
+
+int
+main(int argc, char *argv[])
+{
+	int opt;
+	int do_fork = 1;
+	while ((opt = getopt(argc, argv, "hf")) > 0) {
+		switch (opt) {
+		case 'f':
+			do_fork = 0;
+			break;
+		case 'h':
+			usage(argv[0]);
+			exit(0);
+			break;
+		case '?':
+			usage(argv[0]);
+			exit(-1);
+		}
+	}
+
 #ifndef DEBUG
 	/* Make sure we are root */
 	if (getuid() != 0) {
@@ -588,9 +579,9 @@ main(int UNUSED(argc), char *argv[])
 	dropprivs();
 
 	/* run in the background as a daemon */
-	if (daemon(0, 0)) {
+	if (do_fork && daemon(0, 0)) {
 		syslog(LOG_ERR, "daemon() failed: %m");
-		exit(1);
+		cleanup_exit(1);
 	}
 #endif
 
